/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.gui;

import java.io.PipedOutputStream;
import java.util.Observable;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.util.DialogFrame;
import com.backup.zipsnap.ng.job.Config;
import com.backup.zipsnap.ng.job.Job;

public class JobWrapper extends Observable {
	private Job job;
	private GUI gui;
	private Tab tab;
	public Thread thread = new Thread();
	private JobPanel panel;
	public String command;
	private boolean showJobCompleted;

	public boolean isRunning = false;
	public PipedOutputStream inwriter;
	public static enum msg { STATUS,CONFIG,STDIN,MSG_WARNING,MSG_ERROR,MSG_INFO,SNAPSHOT,FINISHED,JOB_PROGRESS,TASK_PROGRESS,TASK_FILE,
		SUMMARY_READY,CONFIRM_SUMMARY,JOB_CONFIG,JOB_INFO,NO_NEW_ITEMS,TASK_ETA,CONFIRM_OVERWRITE_DELETE,TASK_LABEL,ARCHIVE_INFO,
		CONFIRM_OK,SNAPSHOT_FILE,SNAPSHOT_FILE_READY; }

	/*
	 * create a new JobEngine object
	 */
	public JobWrapper(GUI gui, Tab tab) {
		this.gui = gui;
		this.tab = tab;
		this.panel = tab.jobPanel;
	}

	/*
	 * run the job with the given arguments
	 */
	public void run(final String[] args) {
		/* wait for the old job to complete if still running */
		if (thread.isAlive()) {
			try { thread.join(); } 
			catch (InterruptedException e) { e.printStackTrace(); }
		}
		
		job = new Job(true);
		showJobCompleted = true;
		command = args[1];
		if (command.equals("updateConfig")) args[1] = "config";
		if (command.equals("infoSnapshot")) args[1] = "info";
		if (command.equals("info") || command.equals("config") || command.equals("updateConfig") || command.equals("infoSnapshot")) showJobCompleted = false;
		/* run the job in a new thread */
		thread = new Thread() {
			public void run() {
				String arguments = "";
				for (int i = 0; i < args.length; i++) arguments = arguments+" "+args[i];
				GUI.print("Running job with arguments: "+arguments);
				job.run(tab,args);
			}
		};
		/* start the thread */
		thread.start();
	}
	
	/*
	 * abort the current job
	 */
	@SuppressWarnings("deprecation")
	public void stop() {
		thread.stop();
		callback(msg.STATUS,"finished");
		tab.statusBar.setStatus(StatusBar.status.READY);
	}

	/*
	 * set the job as running and notify all the observers
	 */
	public void setRunning(boolean running) {
		if ((!isRunning && running) || (isRunning && !running)) {
			isRunning = running;
			setChanged();
			notifyObservers();
		}
	}

	/*
	 * handle callbacks from the running job
	 */
	public void callback(msg what, Object obj) {
		/* do nothing if running in cli */
		if (!gui.isInGUI) return;
		
		/* handle job running status */
		if (what == msg.STATUS) {
			String value = (String)obj;
			if (value.equals("started")) {
				setRunning(true);
				tab.statusBar.setStatus(StatusBar.status.RUNNING);
				panel.setStatus(JobPanel.status.RUNNING);
			}
			if (value.equals("finished")) {
				setRunning(false);
				panel.setStatus(JobPanel.status.READY);
			}
		}
		/* handle the configuration once read from the job */
		else if (what == msg.CONFIG) {
			/* adjust the configuration of the job for running within the gui */
			tab.config = (Config)obj;
			Main.settings.progressBarStep = 1;
			Main.settings.logDebugAlways = true;
			Main.settings.listFilesOnSummary = true;
			Main.settings.summaryMaxFiles = 0;
			Main.settings.hideProgressDuringBatch = false;
			Main.settings.printWarningsSummary = false;
			Main.settings.infoAfterBackup = true;
			/* config needs to be updated, apply the changes before saving */
			if (command.equals("updateConfig")) tab.configFrame.applyConfig();
		}
		/* store the stdin handle for controlling the job */
		else if (what == msg.STDIN) {
			inwriter = (PipedOutputStream)obj;
		}
		/* handle info messages */
		else if (what == msg.MSG_INFO) {
			tab.statusBar.setMessage("info", (String) obj);
		}
		/* handle warning messages */
		else if (what == msg.MSG_WARNING) {
			tab.statusBar.setMessage("warning", (String) obj+Main.lineSeparator);
		}			
		/* handle error messages */
		else if (what == msg.MSG_ERROR) {
			tab.statusBar.setMessage("error", (String) obj);
		}
		/* handle snapshot table */
		else if (what == msg.SNAPSHOT) {
			panel.setSnapshot((Object[])obj);
		}
		/* handle job finished status */
		else if (what == msg.FINISHED) {
			tab.statusBar.setLabel("Completed in "+(String)obj);
			if (tab.statusBar.getMessageCount("error") > 0) return;
			else if (tab.statusBar.getMessageCount("warning") > 0) gui.windowsManager.show(new DialogFrame(DialogFrame.type.WARNING,"Job completed with warnings"));
			else if (tab.statusBar.getMessageCount("info") > 0 && showJobCompleted) gui.windowsManager.show(new DialogFrame(DialogFrame.type.INFO,"Job completed succesfully"));
		}
		/* handle job progress updates */
		else if (what == msg.JOB_PROGRESS) {
			panel.setJobProgress((String)obj);
		}
		/* handle task progress updates */
		else if (what == msg.TASK_PROGRESS) {
			panel.setTaskProgress(new Integer((Integer) obj).intValue());
		}
		/* handle task ETA updates */
		else if (what == msg.TASK_ETA) {
			panel.setTaskEta((String)obj);
		}
		/* handle filetree updates for tasks */
		else if (what == msg.TASK_FILE) {
			panel.setTaskFile((String[])obj);
		}
		/* summary is ready to show */
		else if (what == msg.SUMMARY_READY) {
			panel.setSummaryReady();
		}
		/* handle job info messages */
		else if (what == msg.JOB_INFO) {
			panel.clearSnapshotTable();
		}
		/* handle the job configuration settings */
		else if (what == msg.JOB_CONFIG) {
			panel.loadConfig((Config)obj);
		}
		/* handle now new items messages */
		else if (what == msg.NO_NEW_ITEMS) {
			gui.windowsManager.show(new DialogFrame(DialogFrame.type.INFO,"No items need to be added/restored/deleted from the archive"));
			showJobCompleted = false;
			panel.setNoFile();
		}
		/* handle overwrite/delete confirmation */
		else if (what == msg.CONFIRM_OVERWRITE_DELETE) {
			panel.setConfirmOverwriteDelete((DialogFrame) gui.windowsManager.show(new DialogFrame(DialogFrame.type.YES_NO_ALWAYS_NEVER,(String)obj)));
		}
		/* handle summary confirmations */
		else if (what == msg.CONFIRM_SUMMARY) {
			panel.setConfirmSummary((DialogFrame) gui.windowsManager.show(new DialogFrame(DialogFrame.type.YES_NO,(String)obj)));
		}
		/* handle warning confirmation */
		else if (what == msg.CONFIRM_OK) {
			panel.setConfirmWarning((DialogFrame) gui.windowsManager.show(new DialogFrame(DialogFrame.type.OK,(String)obj)));
		}
		/* handle the task panel label */
		else if (what == msg.TASK_LABEL) {
			panel.setTaskPanelLabel((String)obj);
		}
		/* handle archive info message */
		else if (what == msg.ARCHIVE_INFO) {
			panel.setArchiveInfo((String)obj);
		}
		/* handle filetree updates for snapshot info */
		else if (what == msg.SNAPSHOT_FILE) {
			panel.setSnapshotFile((String[])obj);
		}
		/* snapshot info filetree ready to show*/
		else if (what == msg.SNAPSHOT_FILE_READY) {
			panel.setSnapshotFileReady();
		}
	}

}
