/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.NavigableMap;

import net.lingala.zip4j.core.ZipFile;
import net.lingala.zip4j.model.FileHeader;
import net.lingala.zip4j.model.ZipParameters;
import net.lingala.zip4j.util.Zip4jConstants;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.unit.SliceUnit;
import com.backup.zipsnap.ng.job.unit.SnapshotUnit;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.UserIO;


/**
 * Represent a catalog.
 */
public class Catalog
{
	/* snapshots in this catalog */
	public final List<SnapshotUnit> snapshotsIndex = new ArrayList<SnapshotUnit>();
	/* files/directories in this catalog (to be sorted by their name field) */
	public final List<FileUnit> files = new ArrayList<FileUnit>();
	/* file slices */
	public final HashMap<String,SliceUnit> slices = new HashMap<String,SliceUnit>();
	/* catalog file pointers */
	private Pattern catalogNamePattern = null;
	public File catalogFile = null;

	/**
	 * constructor for an empty catalog.
	 */
	public Catalog() {

	}

	/**
	 * Reload the catalog
	 */
	public void reload() {
		/* empty data structures */
		snapshotsIndex.clear();
		files.clear();
		Job.config.get().directories.clear();
		slices.clear();
		/* load the catalog */
		load();
	}

	/**
	 * Load the latest catalog
	 */
	public void load() {
		BufferedReader br = null;
		ZipFile zf = null;
		int numSnapshots = 0;
		int numFiles = 0;
		int numSlices = 0;
		double versionNum = 0;
		catalogNamePattern = Pattern.compile("zipsnap\\."+
				Pattern.quote(Job.config.get().jobName)+ 
				"\\.([0-9]+)\\.([0-9]{8}\\-[0-9]{6})\\.catalog\\.zip");

		catalogFile = getLatestCatalog();

		if (catalogFile == null) {
			/* initialize a new empty catalog */
			UserIO.print(UserIO.out.DEBUG,"Latest catalog file: (none found)");
			return;
		} 

		UserIO.print(UserIO.out.DEBUG,"Latest catalog file: " + catalogFile.getName());
		/* open the catalog file */
		final String catalogFileName = catalogFile.getName();
		if (!catalogFile.exists()) UserIO.reportError("Catalog \"" + catalogFileName + "\" does not exist.");
		try {
			/* read a compressed text file inside a ZIP volume */
			zf = new ZipFile(catalogFile);
			final String zippedFilename = catalogFileName.substring(0, catalogFileName.lastIndexOf("."))+".txt";
			FileHeader fileHeader = zf.getFileHeader(zippedFilename);
			br = new BufferedReader(new InputStreamReader(zf.getInputStream(fileHeader)));
		} catch (Exception e) {
			UserIO.reportError("Unable to open the catalog \"" + catalogFileName + "\": " + UserIO.getExceptionMessage(e));
		}

		/* read and parse each line of the catalog */
		ParseNextLine:
			for (int lineNum = 1; ; lineNum++) {
				String lineString = null;
				/* read the line */
				try {
					if (!br.ready()) break ParseNextLine;
					lineString = br.readLine().trim();
				} catch (Exception e) {
					UserIO.reportError("Unable to read line " + lineNum + " of catalog \"" + catalogFileName + "\": " + UserIO.getExceptionMessage(e));
				}
				/* skip this line if empty or if it begins with '#' */
				if (lineString.isEmpty() || lineString.startsWith("#")) continue ParseNextLine;
				String s = lineString;
				/* parse the line */
				try {
					if (s.startsWith("v:")) {
						/* ZIPSNAP_VERSION:version  */
						/*                    1      */
						final Pattern patternZipSnapVersion = Pattern.compile(Pattern.quote("v:") + "([0-9]+\\.[0-9]+)");
						final Matcher m = patternZipSnapVersion.matcher(s);
						if (!m.matches()) throw new Exception("Malformed VERSION line.");
						versionNum = Double.parseDouble(m.group(1));
						if (versionNum > Main.version) 
							UserIO.reportWarning(UserIO.out.INFO,"Catalog \"" + catalogFileName + "\" was created by a newer version of the program, and may not be read correctly.");
						if (versionNum < 3.0)
							UserIO.reportError("Catalog \"" + catalogFileName + "\" was created by an older version of the program, and may not be read correctly.");
					} 
					else if (s.startsWith("ns:")) {
						/* NUMBER_OF_SNAPSHOTS:number   */
						/*                       1      */
						final Pattern patternNumberOfSnapshots = Pattern.compile(Pattern.quote("ns:") + "([0-9]+)");
						final Matcher m = patternNumberOfSnapshots.matcher(s);
						if (!m.matches()) throw new Exception("Malformed NUMBER_OF_SNAPSHOTS line.");
						numSnapshots = Integer.parseInt(m.group(1));
					} 
					else if (s.startsWith("nf:")) {
						/* NUMBER_OF_FILES:number   */
						/*                   1      */
						final Pattern patternNumberOfFiles = Pattern.compile(Pattern.quote("nf:") + "([0-9]+)");
						final Matcher m = patternNumberOfFiles.matcher(s);
						if (!m.matches()) throw new Exception("Malformed NUMBER_OF_FILES line.");
						numFiles = Integer.parseInt(m.group(1));
					} 
					else if (s.startsWith("s:")) {
						/* SNAPSHOT:number;hostname;time;volumes;isEncrypted;isDeleted  */
						/*            1     2        3       4        5           6 */
						final Pattern patternSnapshot = Pattern.compile(Pattern.quote("s:") + "([0-9]+)" +
								Pattern.quote(";") + "(.+)"+
								Pattern.quote(";") + "([0-9\\-]+)"+
								Pattern.quote(";") +  "([0-9]+)"+
								Pattern.quote(";") +  "([0-1]+)"+
								Pattern.quote(";") +  "([0-1]+)"
								);
						final Matcher m = patternSnapshot.matcher(s);
						if (!m.matches()) throw new Exception("Malformed SNAPSHOT line.");
						final int number = Integer.parseInt(m.group(1));
						final String hostname = m.group(2);
						final String time = m.group(3);
						final int volumes = Integer.parseInt(m.group(4));
						final int encrypted = Integer.parseInt(m.group(5));
						boolean isEncrypted = encrypted == 1 ? true : false;
						final int deleted = Integer.parseInt(m.group(6));
						boolean isDeleted = deleted == 1 ? true : false;

						if (number != (this.snapshotsIndex.size() + 1)) {
							UserIO.reportWarning(UserIO.out.INFO,"Malformed snapshot entry found on line " + lineNum +
									" of catalog \"" + catalogFileName + "\": \"" + lineString + "\" " +
									"Expected snapshot number " + (this.snapshotsIndex.size() + 1) +
									", but found snapshot number " + number);
							continue ParseNextLine;
						}
						this.snapshotsIndex.add(new SnapshotUnit(number,hostname,time,volumes,isEncrypted,isDeleted));
					} 
					else if (s.startsWith("f:") || s.startsWith("d:")) {
						/* FILE:baseDirectory;filePath;size;compressedSize;time;checksum;snapshots;volumes;sliceSize;slices             */
						/*           1           2       3          4       5     6         7          8        9       10   */
						final Pattern patternFile = Pattern.compile(
								"^[f|d]:(.*)" +
										Pattern.quote(";") + "(.*)" +
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") + "([0-9a-fA-F]*)" +
										Pattern.quote(";") + "([0-9\\,]+)"+
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") + "(.*)" +
										Pattern.quote(";")
								);
						final Matcher m = patternFile.matcher(s);
						if (!m.matches()) throw new Exception("Malformed FILE line.");
						/* determine if it is a file or a directory */
						boolean isDirectory = false;
						if (s.startsWith("d:")) isDirectory = true;
						/* neutral and relative pathname of file/directory */
						final String directory = catalogToNeutral(m.group(1));
						final String filePath = m.group(2);
						if (filePath == null|| directory == null) throw new Exception("Malformed file/directory name \"" + m.group(1) + "\" \"" + m.group(2) + "\".");

						/* handle the root directory */
						FileUnit d = null;
						if ((Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.origin) || ! Job.command.get().equals("restore")) {
							/* destination directory will be the original directory */
							d = new FileUnit(FileIO.neutralToNative(directory),"",true);
						} else if (Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.extractInto) {
							/* destination directory will be extractInto directory and the normalized path */
							d = new FileUnit(Job.config.get().extractInto.nativeName+Main.fileSeparator+FileIO.getNormalizedPath(directory),"",true);        
						} else if (Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.mergeDirectory) {
							/* destination directory will be the mapped directory */
							if (Job.config.get().mergeDirectory.containsKey(directory)) {
								d = new FileUnit(Job.config.get().mergeDirectory.get(directory).nativeName,"",true);
							} else {
								/* if there is no valid mapping, restore to the original location */
								d = new FileUnit(FileIO.neutralToNative(directory),"",true);       
							}
						}
						if (!Job.config.get().directories.contains(d)) Job.config.get().directories.add(d);

						/* handle the file */
						FileUnit u = null;
						if ((Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.origin) || ! Job.command.get().equals("restore")) {
							/* destination location will be the original location */
							u = new FileUnit(FileIO.neutralToNative(directory),FileIO.neutralToNative(filePath),isDirectory);
						} else if (Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.extractInto) {
							/* destination location will be the extractInto directory and the normalized path */
							u = new FileUnit(Job.config.get().extractInto.nativeName+Main.fileSeparator+FileIO.getNormalizedPath(FileIO.neutralToNative(directory)),FileIO.neutralToNative(filePath),isDirectory);
							FileUnit restoreFromFile = new FileUnit(FileIO.neutralToNative(directory),FileIO.neutralToNative(filePath),isDirectory);
							/* link the location in the archive where the zip file has been backed up */
							u.restoreFrom = restoreFromFile;
						} else if (Job.command.get().equals("restore") && Job.config.get().restoreMode == Config.restoreType.mergeDirectory) {
							/* destination location will be the mapped directory */
							if (Job.config.get().mergeDirectory.containsKey(directory)) {
								u = new FileUnit(Job.config.get().mergeDirectory.get(directory).nativeName,FileIO.neutralToNative(filePath),isDirectory);
								FileUnit restoreFromFile = new FileUnit(FileIO.neutralToNative(directory),FileIO.neutralToNative(filePath),isDirectory);
								/* link the location in the archive where the zip file has been backed up */
								u.restoreFrom = restoreFromFile;
							} else {
								/* if there is no valid mapping, restore to the original location */
								u = new FileUnit(FileIO.neutralToNative(directory),FileIO.neutralToNative(filePath),isDirectory);
							}
						}
						/* store all the information*/
						u.size = Long.parseLong(m.group(3));
						u.compressedSize = Long.parseLong(m.group(4));
						u.time = Long.parseLong(m.group(5));
						u.checksum = m.group(6);

						/* snapshots that contain this file/directory */
						final List<Integer> snapshotsIndex = new ArrayList<Integer>();
						final String[] tokens = m.group(7).split("\\,");
						for (String t : tokens) snapshotsIndex.add(Integer.parseInt(t) - 1);  /* convert "number" to "index" */
						u.snapshotsIndex.addAll(snapshotsIndex);
						u.volume = Integer.parseInt(m.group(8));
						u.sliceSize = Long.parseLong(m.group(9));

						/* parse the list of checksums */ 
						final String[] checksumList = m.group(10).split("\\,");
						int sliceIndex = 0;
						for (String checksum : checksumList) {
							if (checksum.equals("")) continue;
							SliceUnit slice = new SliceUnit(checksum);
							u.sliceList.add(sliceIndex,slice);
							sliceIndex++;
							/* add the slice to the global slice list */
							if (!slices.containsKey(checksum)) slices.put(checksum,slice);
							/* add the file to to the slice */
							slices.get(checksum).files.add(u);
						}
						u.numSlices = u.sliceList.size();

						/* add the file to the file list */
						this.files.add(u);
					}
					else if (s.startsWith("p:")) {
						/* SLICE:checksum;snapshot;volume;size  */
						/*            1      2        3     4   */
						final Pattern patternSlice = Pattern.compile(
								Pattern.quote("p:") + "([0-9a-fA-F]+)" +
										Pattern.quote(";") + "([0-9]+)" +
										Pattern.quote(";") +  "([0-9]+)" +
										Pattern.quote(";") +  "([0-9]+)" );
						final Matcher m = patternSlice.matcher(s);
						if (!m.matches()) throw new Exception("Malformed SLICE line.");
						String checksum = m.group(1);

						SliceUnit slice = slices.get(checksum);
						if (slice == null) {
							slice = new SliceUnit(checksum);
							slices.put(checksum,slice);
						}
						/* add additional information to the (empty) slice */
						slice.snapshot = Integer.parseInt(m.group(2));
						slice.volume = Integer.parseInt(m.group(3));
						slice.size = Long.parseLong(m.group(4));
					} 
					else if (s.startsWith("np:")) {
						/* NUMBER_OF_SLICES:number   */
						/*                   1      */
						final Pattern patternNumberOfSlices = Pattern.compile(Pattern.quote("np:") + "([0-9]+)");
						final Matcher m = patternNumberOfSlices.matcher(s);
						if (!m.matches()) throw new Exception("Malformed NUMBER_OF_SLICES line.");
						numSlices = Integer.parseInt(m.group(1));
					} 
					else {
						throw new Exception("Unexpected line encountered.");
					}
				} catch (Exception e) {
					UserIO.reportWarning(UserIO.out.DEBUG,"Unable to parse line " + lineNum +
							" of catalog \"" + catalogFileName + "\": \"" + lineString + "\" " +
							UserIO.getExceptionMessage(e) + " This line will be ignored.");
					continue ParseNextLine;
				}
				continue ParseNextLine;
			}

		/* close catalog */
		try {
			br.close();
		} catch (Exception e) {
			UserIO.reportWarning(UserIO.out.INFO,"Unable to close catalog \"" + catalogFileName + "\": " + UserIO.getExceptionMessage(e));
		}

		/* sanity checks */
		if (numSnapshots != this.snapshotsIndex.size())
			UserIO.reportWarning(UserIO.out.INFO,"Inconsistent number of snapshots found in catalog: " +
					"Expected " + numSnapshots + " snapshots, but parsed " + this.snapshotsIndex.size());
		if (numFiles != this.files.size())
			UserIO.reportWarning(UserIO.out.INFO,"Inconsistent number of files found in catalog: " +
					"Expected " + numFiles + " files, but parsed " + this.files.size());
		if (numSlices != this.slices.size())
			UserIO.reportWarning(UserIO.out.INFO,"Inconsistent number of data slices found in catalog: " +
					"Expected " + numSlices + " slices, but parsed " + this.slices.size());      
		/* sort files/directories in catalog */
		Collections.sort(this.files);
		check();
	}


	/**
	 * Write catalog to a file.
	 */
	public void writeToFile(final String catalogFilename) {
		File catalogFileTxt = new File(Job.config.get().archive.file, catalogFilename + ".txt");
		File catalogFileZip = new File(Job.config.get().archive.file, catalogFilename + ".zip");

		UserIO.print(UserIO.out.DEBUG,"Creating new catalog " + catalogFileZip.getName());
		/* create parent directories if necessary */
		FileIO.checkFile(catalogFileTxt);
		FileIO.checkFile(catalogFileZip);
		PrintWriter pw = null;

		/* open the catalog writer */
		try {
			pw = new PrintWriter(catalogFileTxt);
		} catch (Exception e) {
			UserIO.reportError("Unable to create catalog \"" + catalogFileTxt.getName() + "\": " + UserIO.getExceptionMessage(e));
		}
		pw.println("# " + Main.programTitle);
		pw.println("# COMMENT: Lines beginning with a '#' character are ignored by the program");

		/* program version */
		pw.println();
		pw.println("# ZipSnapNG version");
		pw.println("v:" + Main.version);

		/* Snapshots */
		pw.println();
		pw.println("# SNAPSHOT:number;hostname;time;volumes;isEncrypted;isDeleted");

		for (int i = 0; i < this.snapshotsIndex.size(); i++){
			String encrypted = this.snapshotsIndex.get(i).isEncrypted ? "1" : "0";
			String deleted = this.snapshotsIndex.get(i).isDeleted ? "1" : "0";
			pw.println("s:" + (i + 1) + ";" + Job.config.get().hostname+ ";"+this.snapshotsIndex.get(i).time + ";" + 
					this.snapshotsIndex.get(i).volumesIndex+";"+encrypted+";"+deleted);
		}
		pw.println("# Number of snapshots");
		pw.println("ns:" + this.snapshotsIndex.size());

		/* Files */
		pw.println();
		pw.println("# FILE:baseDirectory;filePath;uncompressedSize;compressedSize;time;checksum;snapshots;volumes;sliceSize;slices");

		int skippedFiles = 0;
		for (FileUnit u : this.files) {
			/* skip deleted or invalid items */
			if (u.isDeleted || u.isInvalid) {
				skippedFiles++;
				continue;
			}
			/* determine if it is a file or a directory */
			String fileType = "";
			if (u.isDirectory) fileType = "d:";
			else fileType = "f:";
			pw.print(fileType +
					neutralToCatalog(FileIO.nativeToNeutral(u.directory)) + ";" +
					neutralToCatalog(FileIO.nativeToNeutral(u.filePath)) + ";" +
					u.size + ";" +
					u.compressedSize + ";" +
					u.time + ";" +
					u.checksum + ";");
			/* print the snapshots */
			for (int i = 0; i < u.snapshotsIndex.size() - 1; i++) pw.print((u.snapshotsIndex.get(i) + 1) + ","); 
			pw.print(u.snapshotsIndex.get(u.snapshotsIndex.size() - 1) + 1); 
			pw.print(";" + u.volume + ";");
			pw.print(u.sliceSize + ";");
			for (int i = 0; i < u.sliceList.size(); i++) {
				pw.print(u.sliceList.get(i).checksum); 
				if (i != u.sliceList.size()-1) pw.print(",");
			}
			pw.println(";");
		}

		pw.println("# Number of files/directories");
		pw.println("nf:" + (this.files.size()-skippedFiles));

		pw.println();
		int skippedSlices = 0;
		pw.println("# SLICE:checksum;snapshot;volume;size");
		for ( String checksum : slices.keySet() ) {
			SliceUnit slice = slices.get(checksum);
			/* skip deleted or invalid items */
			if (slice.isInvalid) {
				skippedSlices++;
				continue;
			}
			if (slice.snapshot == -1 || slice.volume == -1) {
				UserIO.reportWarning(UserIO.out.INFO,"The slice "+slice.checksum+" is invalid and will not be saved");
				skippedSlices++;
				continue;
			}
			pw.println("p:"+slice.checksum+";"+slice.snapshot+";"+slice.volume+";"+slice.size);
		}

		pw.println("# Number of data slices");
		pw.println("np:" + (this.slices.size()-skippedSlices));

		pw.println();
		pw.println("# End-of-Catalog");

		/* close the catalog writer */
		pw.flush();
		pw.close();

		/* check for errors */
		if (pw.checkError()) UserIO.reportError("An error was encountered while creating catalog \"" +catalogFileTxt.getName());

		/* create the zip file */
		try {
			FileIO.checkFile(catalogFileZip);
			ZipFile zipFile = new ZipFile(catalogFileZip.getPath());
			ZipParameters parameters = new ZipParameters();
			parameters.setCompressionMethod(Zip4jConstants.COMP_DEFLATE);
			parameters.setCompressionLevel(Zip4jConstants.DEFLATE_LEVEL_NORMAL);
			ArrayList<File> filesToAdd = new ArrayList<File>();
			filesToAdd.add(catalogFileTxt);
			zipFile.addFiles(filesToAdd,parameters);
		} catch (Exception e) {
			UserIO.reportError("Unable to create the catalog ZIP file: " + UserIO.getExceptionMessage(e));
		}
		/* set timestamp of catalog file */
		final FileIO.FileIOResult result = FileIO.setFileTime(catalogFileZip, Job.currentTime.get().getTime());
		if (!result.success) UserIO.reportWarning(UserIO.out.INFO,"Unable to set last-modified time of catalog \"" + catalogFileZip.getName() + "\".");
		/* remove the tmp txt file */
		FileIO.FileIOResult resultDelete = FileIO.deleteFile(catalogFileTxt);
		if (!resultDelete.success)  {
			UserIO.reportWarning(UserIO.out.INFO,"Unable to delete catalog temporary file "+catalogFileTxt.getPath()+": "+ result.errorMessage);
		}
	}

	/**
	 * Convert a file name from neutral form to catalog form.
	 */
	private static String neutralToCatalog(final String path) {
		/* return value */
		final StringBuilder t = new StringBuilder();
		/* escape sequences '\n' and '\r' */
		for (int i = 0; i < path.length(); i++) {
			final char c = path.charAt(i);
			if (c == '\n') t.append("\\n");
			else if (c == '\r') t.append("\\r");
			else if (c == '\\') t.append("\\\\");
			else t.append(c);
		}
		return t.toString();
	}


	/**
	 * Convert a file/directory name from catalog form to neutral form.
	 */
	private static String catalogToNeutral(final String path) {
		/* return value */
		final StringBuilder t = new StringBuilder();
		/* escape sequences '\n' and '\r' */
		for (int i = 0; i < path.length(); i++) {
			final char c = path.charAt(i);
			if (c == '\\') {
				if (i + 1 < path.length()) {
					/* read next character */
					i++;
					final char d = path.charAt(i);
					if (d == 'n') t.append('\n');
					else if (d == 'r') t.append('\r');
					else if (d == '\\') t.append('\\');
					else return null;
				} else {
					return null;
				}
			} else {
				t.append(c);
			}
		}
		return t.toString();
	}

	/**
	 * find the best match of a given file unit in the catalog
	 */  
	public FileUnit find(FileUnit u, int i) {
		/* index of the best match */
		int bestMatch = -1;
		int bestMatchLatestSnapshot = -1;
		List<FileUnit> candidates = new ArrayList<FileUnit>();
		/* match files upwards */
		MatchNextFileUp:
			for (int j = i - 1; j >= 0; j--) {
				final FileUnit c = Job.catalog.get().files.get(j);
				/* if the name does not match we went too far */
				if (!u.name.equals(c.name)) break MatchNextFileUp;
				c.line = j;
				candidates.add(c);
			}
		/* match files downwards */
		MatchNextFileDown:
			for (int j = i; j < Job.catalog.get().files.size(); j++) {
				final FileUnit c = Job.catalog.get().files.get(j);
				/* if the name does not match we went too far */
				if (!u.name.equals(c.name)) break MatchNextFileDown;
				c.line = j;
				candidates.add(c);
			}

			/* for all the candidates which have the same name */
			MatchNext:
				for (FileUnit c : candidates) {
					/* name is already matched, next match size, time, and checksum */
					if ((u.size != c.size) || (Math.abs(u.time - c.time) > Main.settings.timeTolerance) || (! u.checksum.equals("") && !u.checksum.equals(c.checksum))) {
						UserIO.print(UserIO.out.ALL," not matching ("+
								u.size+"!="+ c.size+
								" || "+u.time+" - "+c.time+" > "+Main.settings.timeTolerance+
								" || "+u.checksum+" != "+c.checksum+") ",false);
						continue MatchNext;
					}
					/* match found: proceed to compare latest snapshot indices, if necessary */
					final int latestSnapshot = Collections.max(c.snapshotsIndex)+1;
					UserIO.print(UserIO.out.VERBOSE," matching snaphot "+latestSnapshot+" ",false);
					if ((bestMatch < 0) || (latestSnapshot > bestMatchLatestSnapshot)) {
						bestMatch = c.line;
						bestMatchLatestSnapshot = latestSnapshot;
					}
				}
			/* we have a match */
			if (bestMatch >= 0) {
				return Job.catalog.get().files.get(bestMatch);
			} else return null;
	}

	/**
	 * check if the catalog has a given shapshot number
	 */  
	public boolean hasSnapshot(int snapshot) {
		for (SnapshotUnit s : snapshotsIndex) {
			if (s.number == snapshot) return true;
		}
		return false;
	}

	/**
	 * return all the catalog files in the archive directory
	 */  
	public List<File> getCatalogFiles() {
		final List<File> catalogFiles = new ArrayList<File>();
		/* get contents of the archive directory */
		final File dirContents[] = Job.config.get().archive.file.listFiles();
		if (dirContents == null) UserIO.reportError("Unable to get contents of the specified archive directory \"" + Job.config.get().archive.nativeName);
		/* create regex pattern to match files */
		for (File f : dirContents) {
			if (f.isFile()) {
				final Matcher m = catalogNamePattern.matcher(f.getName());
				if (m.find()) {
					catalogFiles.add(f);
				}
			}
		}
		return catalogFiles;
	}


	/**
	 * get the latest catalog
	 */  
	public File getLatestCatalog() {
		/* select the latest catalog file using the following sequence of "keys": */
		/*   1. Largest snapshot number (value of n)                              */
		/*   2. Latest last-modified time                                         */
		final NavigableMap<Integer,File> candidateCatalogs = new TreeMap<Integer,File>();
		for (File u : getCatalogFiles()) {
			final Matcher m = catalogNamePattern.matcher(u.getName());
			if (m.matches()) {
				/* found a match, i.e. this file is a catalog; get value of n */
				int snapshot = Integer.parseInt(m.group(1));
				final File c = candidateCatalogs.get(snapshot);
				if ((c == null) || (u.lastModified() >= c.lastModified())) {
					/* add this catalog file as a candidate */
					candidateCatalogs.put(snapshot, u);
				}
			}
		}
		/* check if there are any candidate catalog files */
		if (candidateCatalogs.isEmpty()) return null;
		/* return catalog file corresponding to highest snapshot number */
		return candidateCatalogs.get(candidateCatalogs.lastKey());
	}

	/**
	 * remove all but the recent catalog file
	 */  
	public void purgeCatalogFiles() {
		List<File> catalogFiles = getCatalogFiles();
		File latestCatalog = getLatestCatalog();
		catalogFiles.remove(latestCatalog);
		if (catalogFiles.size() > 0) {
			for (File catalogFile : catalogFiles) {
				UserIO.print(UserIO.out.DEBUG,"Deleting catalog file "+catalogFile.getName());
				FileIO.FileIOResult result = FileIO.deleteFile(catalogFile);
				if (!result.success)  {
					UserIO.reportWarning(UserIO.out.INFO,"Unable to delete catalog file "+catalogFile.getPath()+": "+ result.errorMessage);
				}
			}
		}
	}

	/**
	 * Check if the catalog is valid and contains snapshots 
	 */
	public void isValid() {
		String msg = "";
		/* check if a catalog is found */
		if (catalogFile == null) msg = "Unable to find a catalog in the specified archive directory " + Job.config.get().archive.nativeName;
		/* check that catalog contains at least one snapshot */
		if (catalogFile != null && snapshotsIndex.isEmpty()) msg = "No snapshots found in the catalog \"" + catalogFile.getName() + "\" for this archive";
		if (! msg.equals("")) {
			UserIO.print(UserIO.out.INFO,msg);
			Job.exit(0);
		}
	}

	/**
	 * Perform an integrity check of the catalog
	 */
	@SuppressWarnings("unchecked")
	public void check() {
		UserIO.print(UserIO.out.DEBUG,"Checking catalog integrity...",false);
		/* check files integrity */
		for (FileUnit u : files) {
			long calculatedSize = 0;
			for (SliceUnit slice : u.sliceList) {
				if (! slices.containsKey(slice.checksum)) 
					UserIO.reportWarning(UserIO.out.INFO,"The file "+u.nativeName+" is referencing an invalid slice " + slice.checksum);
				else calculatedSize += slices.get(slice.checksum).size;
			}
			if (calculatedSize > 0 && u.size != calculatedSize) 
				UserIO.reportWarning(UserIO.out.INFO,"The file "+u.nativeName+" has an inconsisted size (size: " + u.size+" calculated: "+calculatedSize+")");
		}

		/* check ZIP archives integrity */
		if (Main.settings.catalogDeepCheck) {
			UserIO.print(UserIO.out.DEBUG,"files...",false);
			ProcessNextSnapshot:
				for (int snapshotIndex = 0; snapshotIndex < snapshotsIndex.size(); snapshotIndex++) {
					ProcessNextVolume:
						for (int volume = 0; volume <= snapshotsIndex.get(snapshotIndex).volumesIndex; volume++) {
							/* build the volume filename */
							String volumeName = FileIO.getCustomFilename("archive",(snapshotIndex + 1),Job.catalog.get().snapshotsIndex.get(snapshotIndex).time);
							if (volume > 0) volumeName = volumeName+ "."+volume;

							/* open the zip file */
							File zipFile = new File(Job.config.get().archive.file, volumeName + ".zip");
							if (zipFile == null || !zipFile.exists()) {
								UserIO.reportError("Unable to locate volume " + zipFile.getPath());
								continue ProcessNextVolume;
							}
							ZipFile zf = null;
							try {
								zf = new ZipFile(zipFile);
							} catch (Exception e) {
								UserIO.reportError("Unable to open ZIP volume " + zipFile.getPath()+ ": " + UserIO.getExceptionMessage(e));
								continue ProcessNextVolume;
							}

							/* Get the list of file headers from the zip file */
							List<FileHeader> fileHeaderList = new ArrayList<FileHeader>();
							try {
								fileHeaderList = zf.getFileHeaders();
							} catch (Exception e) {
								if (UserIO.getExceptionMessage(e).contains("Negative seek offset")) UserIO.reportWarning(UserIO.out.DEBUG,"Skipping ZIP volume " + zipFile.getPath()+ " which may be empty (" + UserIO.getExceptionMessage(e)+")");
								else UserIO.reportError("Unable to retrieve the list of files of the ZIP volume " + zipFile.getPath()+ ": " + UserIO.getExceptionMessage(e));
								continue ProcessNextVolume;
							}
							for (int i = 0; i < fileHeaderList.size(); i++) {
								FileHeader fileHeader = (FileHeader)fileHeaderList.get(i);
								String name = fileHeader.getFileName();
								if (slices.containsKey(name)) {
									SliceUnit slice = slices.get(name);
									/* set slice as stored */
									slice.isInZip = true;
									/* check if the size matches */
									long size = fileHeader.getUncompressedSize();
									if (slice.size != size) 
										UserIO.reportWarning(UserIO.out.INFO,"The slice "+name+" contained in " + zipFile.getName() + " has an inconsitent size (catalog: "+slice.size+", archive: "+size+")");
								} else {
									UserIO.reportWarning(UserIO.out.INFO,"The slice "+name+" contained in " + zipFile.getName() + " is not available in the catalog");
								}
							}
							continue ProcessNextVolume;
						}
			continue ProcessNextSnapshot;
				}
		}
		UserIO.print(UserIO.out.DEBUG,"archives...",false);

		/* check slices integrity */
		for (SliceUnit slice : slices.values()) {
			if (slice.files.size() == 0) 
				UserIO.reportWarning(UserIO.out.INFO,"The slice " + slice.checksum + " is not referenced by any file");
			if (Main.settings.catalogDeepCheck && !slice.isInZip) 
				UserIO.reportWarning(UserIO.out.INFO,"The slice " + slice.checksum + " is not stored in any ZIP archive");
		}
		UserIO.print(UserIO.out.DEBUG,"slices...",false);
	}


}
