/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.backup.zipsnap.ng.job;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.lingala.zip4j.core.ZipFile;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.JobWrapper;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.unit.SliceUnit;
import com.backup.zipsnap.ng.job.unit.SnapshotUnit;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.ProgressBar;
import com.backup.zipsnap.ng.job.util.UserIO;

/**
 * Maintain an archive 
 */
public class Purge
{

	/**
	 * Purge an archive
	 */
	private static String purgedString = "";
	
	public static void doPurge(boolean directCall) {
		if (directCall) {
			UserIO.print(UserIO.out.INFO,"Requested action: PURGE ARCHIVE");
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[INITIALIZING]");
			if (directCall) Job.gui(JobWrapper.msg.JOB_PROGRESS,"INITIALIZING");
			/****************************************
			 * (1) CHECK CATALOG FOR THE ARCHIVE    *
			 ****************************************/

			/* check if a catalog is found */
			Job.catalog.get().isValid();

			/*********************************
			 * (2) CHECK RETENTION POLICIES *
			 *******************************/
			boolean purgeConfigured = isPurgeConfigured();

			if (! purgeConfigured) {
				UserIO.print(UserIO.out.INFO,"No purging options configured");
				if (directCall) {
					Job.gui(JobWrapper.msg.NO_NEW_ITEMS, "");
					Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGE SUMMARY");
					Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGING THE ARCHIVE");
					Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
					Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");
				}
				return;
			}
		}

		/************************************
		 * (3) IDENTIFY SNAPSHOTS TO PURGE *
		 ***********************************/
		UserIO.print(UserIO.out.INFO,"Identifying snapshots to purge...",false);
		List<Integer> deleteSnapshotsIndex = new ArrayList<Integer>();
		for (SnapshotUnit snapshot : Job.catalog.get().snapshotsIndex) {
			/* ignore already deleted snapshots */
			if (snapshot.isDeleted) continue;
			long secondsBetween = (Job.currentTime.get().getTime() - snapshot.date.getTime()) / 1000;
			long daysBetween = secondsBetween/3600/24;
			//long daysBetween = secondsBetween;
			/* identify snapshots older than the threshold */
			UserIO.print(UserIO.out.VERBOSE,"Snapshot "+snapshot.number+" is "+daysBetween+" days old ");
			if (daysBetween >= Job.config.get().purgeAfterDays) {
				/* add them to the candidate list */
				UserIO.print(UserIO.out.DEBUG,"#"+snapshot.number+" ("+daysBetween+" days old)...",false);
				deleteSnapshotsIndex.add(snapshot.number-1);
				Job.catalog.get().snapshotsIndex.get(snapshot.number-1).isDeleted = true;
			}
		}
		UserIO.print(UserIO.out.INFO,"done");

		/******************************************
		 * (4) PRINT THE SUMMARY                 *
		 *****************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[PURGE SUMMARY]");
		if (directCall) Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGE SUMMARY");
		if (deleteSnapshotsIndex.size() == 0) {
			UserIO.print(UserIO.out.INFO,"No snapshots needs to be purged from the archive");
			if (directCall) {
				Job.gui(JobWrapper.msg.NO_NEW_ITEMS, "");
				Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGING THE ARCHIVE");
				Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
				Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");
			}
			return;
		}
		String purgeSummaryText = "";
		UserIO.print(UserIO.out.INFO,"The following snapshots and associated files will purged from the archive:");
		for (Integer snapshotIndex : deleteSnapshotsIndex) {
			SnapshotUnit snapshot = Job.catalog.get().snapshotsIndex.get(snapshotIndex.intValue());
			UserIO.print(UserIO.out.INFO,"- #"+snapshot.number+" ("+snapshot.date.toString()+")");
			if (directCall) Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following snapshots and associated files will purged from the archive:","#"+snapshot.number+" ("+snapshot.date.toString()+")",""});
			else purgeSummaryText = purgeSummaryText+Main.lineSeparator+ "- #"+snapshot.number+" ("+snapshot.date.toString()+")"; 
		}
		if (directCall) Job.gui(JobWrapper.msg.SUMMARY_READY, "");
		
		/* wait for user confirmation */
		if (Job.config.get().interactive) {
			Job.timer.get().pause("JOB");
			String purgeSummaryQuestion = directCall ? "Do you want to proceed by deleting the snapshots below?"  : "Do you want to proceed by deleting the following snapshots?"+purgeSummaryText;
			char choice = UserIO.userCharPrompt(UserIO.choice.YES_NO,purgeSummaryQuestion);
			Job.timer.get().resume("JOB");
			if (choice == 'N') return;
		}

		/******************************************
		 * (5) REMOVE OLD FILES FROM THE CATALOG *
		 *****************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[PURGING THE ARCHIVE]");
		if (directCall) Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGING THE ARCHIVE");
		UserIO.print(UserIO.out.INFO,"Removing expired files from the catalog...",false);
		int deletedFiles = 0;
		for (int i = 0; i < Job.catalog.get().files.size(); i++) {
			FileUnit u = Job.catalog.get().files.get(i);
			String snapshots = "";
			for (int snapshotsIndex : u.snapshotsIndex) snapshots += (snapshotsIndex+1)+" ";
			u.snapshotsIndex.removeAll(deleteSnapshotsIndex);
			UserIO.print(UserIO.out.ALL,"Evaluating "+u.nativeName+" snapshot "+snapshots);
			if (u.snapshotsIndex.size() == 0) {
				/* if the file is in no snapshots */
				UserIO.print(UserIO.out.DEBUG,"Deleting "+u.nativeName+" snapshot "+snapshots);
				u.isDeleted = true;
				deletedFiles++;
				/* unreference this file from each of its slice */
				for (SliceUnit slice : u.sliceList) {
					UserIO.print(UserIO.out.VERBOSE,"Unreferencing slice "+slice.checksum);
					Job.catalog.get().slices.get(slice.checksum).files.remove(u);
				}
			}
		}
		UserIO.print(UserIO.out.INFO,deletedFiles+" items");
		purgedString = "Purged "+deletedFiles+" items";

		/******************************************
		 * (5) REMOVE SLICES WITHOUT A REFERENCE *
		 *****************************************/
		purgeInvalidSlices();
		Job.gui(JobWrapper.msg.TASK_LABEL,purgedString);

		/*********************************************
		 * (6) WRITE NEW CATALOG FOR THE ARCHIVE     *
		 *********************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[FINALIZING]");
		if (directCall) Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
		Collections.sort(Job.catalog.get().files);

		UserIO.print(UserIO.out.INFO,"Writing the updated catalog to file...",false);
		Job.catalog.get().writeToFile(FileIO.getCustomFilename("catalog"));
		UserIO.print(UserIO.out.INFO,"done");

		if (Main.settings.purgeOldCatalogFiles) {
			UserIO.print(UserIO.out.INFO,"Purging old catalog files...",false);
			Job.catalog.get().purgeCatalogFiles();
			UserIO.print(UserIO.out.INFO,"done");
		}
	    /* by reloading the catalog an integrity check is performed */
	    UserIO.print(UserIO.out.INFO,"Reloading the catalog...",false);
	    Job.catalog.get().reload();
	    UserIO.print(UserIO.out.INFO,"done");
	    
		/*******************************************
		 * (7) SHOW ARCHIVE INFO                 *
		 *****************************************/
		if (directCall && Main.settings.infoAfterBackup) {
			Job.config.get().filter = null;
			Job.gui(JobWrapper.msg.JOB_INFO,"");
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[ARCHIVE SUMMARY]");
			Job.timer.get().start("ARCHIVE SUMMARY");
			Info.archiveSummary(false);
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("ARCHIVE SUMMARY"));
		}
	    
	    if (directCall) Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");

	}

	/**
	 * Check if purging options are configured
	 */
	public static boolean isPurgeConfigured() {
		if (Job.config.get().purgeAfterDays > 0) return true;
		return false;
	}



	/**
	 * Remove from an archive a list of slices
	 */
	public static void purgeInvalidSlices() {
		Job.timer.get().start("purgeInvalidSlices");
		UserIO.print(UserIO.out.INFO,"Optimizing ZIP archives...",false);
		int deleteSlicesCount = 0;
		long deleteSlicesSize = 0;
		/* Map for each snapshot/volume the slices to be deleted */
		/* deleteSlices[snapshot][volume] = SliceUnit */
		final ArrayList<ArrayList<ArrayList<SliceUnit>>> deleteSlices = new ArrayList<ArrayList<ArrayList<SliceUnit>>>(Job.catalog.get().snapshotsIndex.size());
		/* create the data structure first */
		for (int i = 0; i < Job.catalog.get().snapshotsIndex.size(); i++) {
			deleteSlices.add(i, new ArrayList<ArrayList<SliceUnit>>());
			for (int j = 0; j <= Job.catalog.get().snapshotsIndex.get(i).volumesIndex; j++) {
				deleteSlices.get(i).add(j,new ArrayList<SliceUnit>(Job.catalog.get().snapshotsIndex.get(i).volumesIndex+1));
			}
		}

		/* add the slices to delete */
		for (SliceUnit slice : Job.catalog.get().slices.values()) {
			if (slice.files.size() == 0) {
				/* this slice is not referenced by any file */
				if (slice.snapshot == -1 || slice.volume == -1) continue;
				UserIO.print(UserIO.out.ALL,"Marking "+slice.checksum+" to delete from "+slice.snapshot+":"+slice.volume);
				deleteSlices.get(slice.snapshot-1).get(slice.volume).add(slice);
				deleteSlicesCount++;
				deleteSlicesSize += slice.size; 
			}
		}

		ProgressBar progressBar = new ProgressBar();
		long progressDeletedSize = 0;
		DeleteNextSnapshot:
			/* for each snapshot */
			for (int snapshotIndex = 0; snapshotIndex < Job.catalog.get().snapshotsIndex.size(); snapshotIndex++) {
				final ArrayList<ArrayList<SliceUnit>> deleteSlicesSnapshot = deleteSlices.get(snapshotIndex);
				/* for each volume file */
				DeleteNextVolume:
					for (int volume = 0; volume < deleteSlicesSnapshot.size(); volume++) {
						/* files to be extracted for the given volume file*/
						final ArrayList<SliceUnit> volumeSlices = deleteSlicesSnapshot.get(volume);
						if (volumeSlices.isEmpty()) continue DeleteNextVolume;

						/* build the volume filename */
						String volumeName = FileIO.getCustomFilename("archive",(snapshotIndex + 1),Job.catalog.get().snapshotsIndex.get(snapshotIndex).time);
						if (volume > 0) volumeName = volumeName+ "."+volume;

						/* open the zip file */
						File zipFile = new File(Job.config.get().archive.file, volumeName + ".zip");
						if (zipFile == null || !zipFile.exists()) {
							UserIO.reportError("Unable to locate volume " + zipFile.getPath());
							continue DeleteNextVolume;
						}

						/* delete the slices from the volume */
						ZipFile zf = null;
						try {
							zf = new ZipFile(zipFile);
						} catch (Exception e) {
							UserIO.reportError("Unable to open ZIP volume " + zipFile.getPath()+ ": " + UserIO.getExceptionMessage(e));
						}
						/* if zip file is password protected then set the password */
						try {
							if (zf.isEncrypted()) {
								if (Job.config.get().password == null) UserIO.reportError("The ZIP volume " + zipFile.getPath()+ " is encypted but no password is provided");
								zf.setPassword(Job.config.get().password);
							}
						} catch (Exception e) {
							UserIO.reportWarning(UserIO.out.INFO,"Unable to determine if the ZIP volume " + zipFile.getPath()+ " is password protected: " + UserIO.getExceptionMessage(e));
						}

						/* for each file supposed to be extracted from the volume */
						UserIO.print(UserIO.out.DEBUG,"Purging "+zipFile.getName()+"...",false);
						for (SliceUnit slice : volumeSlices) {
							try {
								UserIO.print(UserIO.out.VERBOSE,slice.checksum+"...",false);
								zf.removeFile(slice.checksum);
								Job.catalog.get().slices.remove(slice.checksum);
								progressDeletedSize += slice.size;
								progressBar.display(progressDeletedSize,deleteSlicesSize);
							} catch (Exception e) {
								UserIO.reportWarning(UserIO.out.INFO,"Unable to delete slice "+slice.checksum+" from ZIP volume " + zipFile.getPath()+ ": " + UserIO.getExceptionMessage(e));
							}
						} // end for each slice
						UserIO.print(UserIO.out.DEBUG,"done");
						continue DeleteNextVolume;
					} // end for each volume
				continue DeleteNextSnapshot;
			} // end for each snapshot
		UserIO.print(UserIO.out.INFO,"deleted "+deleteSlicesCount+" data slices ("+FileIO.getSize(deleteSlicesSize)+")");
		purgedString = purgedString +" "+deleteSlicesCount+" data slices ("+FileIO.getSize(deleteSlicesSize)+")";
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("purgeInvalidSlices"));    
	}

}