/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job.unit;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.job.Job;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.UserIO;

/**
 * Represent a file/directory with all its attributes.
 */
public class FileUnit implements Comparable<FileUnit>
{
	/* path and directories */
	public String name;
	public String nativeName;
	public String directory;
	public String filePath;

	/* other file attributes */
	public File file;
	public boolean isDirectory;
	public long size = 0;
	public String checksum = "";
	public long time = 0;

	/* file size */
	public long compressedSize = 0;
	public long partialSize = 0;   /* size of the new slices */

	/* snapshots and volume that contain this file/directory */
	public List<Integer> snapshotsIndex = new ArrayList<Integer>();
	public int volume = 0;

	/* slices containing the file contents */
	public ArrayList<SliceUnit> sliceList = new ArrayList<SliceUnit>();
	public int numSlices = 0;
	public int newSlices = 0;
	public long sliceSize = Main.settings.sliceSize;

	/* other temporary attributes */
	public int line = -1;
	public FileUnit restoreFrom;
	public boolean isDeleted = false;
	public boolean isInvalid = false;

	/**
	 * instantiate a FileUnit by providing the base directory, the filepath
	 */
	public FileUnit(String nativeDirectory, String nativeFilePath,boolean isDirectory, boolean shadowAllow) {
		this.directory = FileIO.trimTrailingSeparator(nativeDirectory);
		this.filePath = FileIO.trimLeadingSeparator(nativeFilePath);
		this.isDirectory = isDirectory;
		if (this.filePath.equals("")) this.nativeName = this.directory;
		else this.nativeName = this.directory+Main.fileSeparator+this.filePath;
//GUI.print(Job.command.get().equals("backup")+"");
//GUI.print(Job.config.get().shadowDevice);
		if (shadowAllow && Job.command.get().equals("backup") && Job.config.get().shadowDevice != null) {
			String shadowName = nativeName.replaceFirst("^\\w:",Matcher.quoteReplacement(Job.config.get().shadowDevice));
			this.file = new File(shadowName);
		} else this.file = new File(nativeName);

		/* set the nautral name and the file object */
		this.name = FileIO.nativeToNeutral(this.nativeName);
	}

	public FileUnit(
			String nativeDirectory,
			String nativeFilePath,
			boolean isDirectory)
	{
		this(nativeDirectory,nativeFilePath,isDirectory,true);
	}

	/**
	 * Comparable object
	 */
	@Override
	public int compareTo(final FileUnit o) {
		return this.name.compareTo(o.name);
	}

	@Override
	public boolean equals(final Object o) {
		if (o instanceof FileUnit) return this.name.equals(((FileUnit) o).name);
		return false;
	}

	@Override
	public int hashCode()  {
		return this.name.hashCode();
	}


	/**
	 * Return true if this file/directory matches the defined filter
	 */
	public boolean matchesFilter() {
		if (Job.config.get().filter == null) return true;
		return Job.config.get().filter.matcher(this.nativeName).matches();
	}

	/**
	 * Calculate and update the checksum of the FileUnit
	 */ 
	public void updateChecksum() {
		/* all the slices will have the same size */
		this.sliceSize = Main.settings.sliceSize;
		/* calculate the checksum if the file exists */
		FileIO.ChecksumResult checksumResult;
		if (file.exists()) checksumResult = FileIO.calculateChecksum(this);
		else {
			UserIO.reportWarning(UserIO.out.DEBUG,"Unable to compute the checksum: the file does not exist anymore");
			UserIO.reportWarning(UserIO.out.INFO,"Skipping " + this.file.getPath());
			isInvalid = true;
			return;
		}
		if (! checksumResult.success) {
			UserIO.reportWarning(UserIO.out.DEBUG,"Unable to compute the checksum: "+checksumResult.errorMessage);
			UserIO.reportWarning(UserIO.out.INFO,"Skipping " + this.file.getPath());
			isInvalid = true;
			return;
		}
		/* record the file checksum first */
		this.checksum = checksumResult.checksum;
		/* create the slices */
		for (int i = 0; i < checksumResult.sliceList.size(); i++) {
			/* add it to the sliceList */
			this.sliceList.add(i,checksumResult.sliceList.get(i));
			this.numSlices++;
		}
		if (partialSize > size) {
			UserIO.reportWarning(UserIO.out.INFO,"Inconsistent size for the file "+nativeName+" which may result corrupted");
			UserIO.reportWarning(UserIO.out.DEBUG,"The file "+nativeName+" has a partialSize greater than its size ("+partialSize+" > "+size+")");
			isInvalid = true;
		}
	}

	/**
	 * return the percentage of the file that is new
	 */  
	public long percentageNew() {
		return FileIO.percentage(this.partialSize,this.size);
	}

	/**
	 * update the file properties (timestamp and size) from the filesystem
	 */    
	public void update() {
		this.time = file.lastModified();
		if (this.isDirectory) this.size = 0L;
		else this.size = this.file.length();
	}

	/**
	 * Set a file as invalid
	 */    
	public void setInvalid() {
		UserIO.print(UserIO.out.VERBOSE,"Marking "+this.nativeName+" as invalid");
		this.isInvalid = true;
		for (SliceUnit slice : sliceList) {
			if (slice.files.contains(this)) slice.files.remove(this);
			if (slice.isNew) { 
				Job.catalog.get().slices.get(slice.checksum).isInvalid = true;
				UserIO.print(UserIO.out.VERBOSE,"Marking slice "+slice.checksum+" as invalid");
			}
		}
	}

}



